from scipy import spatial
import numpy as np


def make_tree(d1=None, d2=None, d3=None):
    active_dimensions = [
        dimension for dimension in [d1, d2, d3] if dimension is not None
    ]
    assert len(active_dimensions) > 0, "Must have at least 1-dimension to make tree"
    if len(active_dimensions) == 1:
        points = np.c_[active_dimensions[0].ravel()]
    elif len(active_dimensions) == 2:
        points = np.c_[active_dimensions[0].ravel(), active_dimensions[1].ravel()]
    else:
        points = np.c_[
            active_dimensions[0].ravel(),
            active_dimensions[1].ravel(),
            active_dimensions[2].ravel(),
        ]
    return spatial.cKDTree(points), len(active_dimensions)


def calculate_overlap(
    point_of_interest, bounding_size, score_radius, sample_shape, dimensions
):
    if dimensions == 1:
        d = point_of_interest[0][0]
        if d <= abs(score_radius - bounding_size):
            vol = score_radius
        elif d >= score_radius + bounding_size:
            vol = 0
        else:
            vol = max(
                0,
                min(bounding_size, d + score_radius)
                - max(-bounding_size, d - score_radius),
            )
    elif sample_shape == "circle":
        if dimensions == 2:
            d = np.sum((np.array(point_of_interest)) ** 2) ** 0.5
            if d <= abs(score_radius - bounding_size):
                vol = np.pi * score_radius**2
            elif d >= score_radius + bounding_size:
                vol = 0
            else:
                r2, R2, d2 = score_radius**2, bounding_size**2, d**2
                r, R = score_radius, bounding_size
                alpha = np.arccos((d2 + r2 - R2) / (2 * d * r))
                beta = np.arccos((d2 + R2 - r2) / (2 * d * R))
                vol = (
                    r2 * alpha
                    + R2 * beta
                    - 0.5 * (r2 * np.sin(2 * alpha) + R2 * np.sin(2 * beta))
                )
        elif dimensions == 3:
            d = np.sum((np.array(point_of_interest)) ** 2) ** 0.5
            if d >= score_radius + bounding_size:
                vol = 0
            elif bounding_size > (
                d + score_radius
            ):  # scoring radius is entirely contained in nucleus
                vol = (4 / 3) * np.pi * score_radius**3
            else:
                vol = (
                    np.pi
                    * (score_radius + bounding_size - d) ** 2
                    * (
                        d**2
                        + (
                            2 * d * score_radius
                            - 3 * score_radius**2
                            + 2 * d * bounding_size
                            - 3 * bounding_size**2
                        )
                        + 6 * score_radius * bounding_size
                    )
                ) / (12 * d)
    elif sample_shape == "rectangle":
        if dimensions == 2:
            pass
        elif dimensions == 3:
            pass

    assert vol > 0, (
        "Attempted to boundary correct a point not within the sample. "
        "Check sample_size and points."
    )
    return vol


def calculate_ripley(
    radii,
    sample_size,
    d1=None,
    d2=None,
    d3=None,
    sample_shape="circle",
    boundary_correct=False,
    CSR_Normalise=False,
):
    results = []
    tree, dimensions = make_tree(d1=d1, d2=d2, d3=d3)
    if type(radii) is not list:
        radii = [radii]
    for radius in radii:
        if dimensions == 1:
            score_vol = radius * 2
            bound_size = sample_size
            counts = 0
            for x in zip(d1):
                if boundary_correct:
                    vol = calculate_overlap(
                        [x], sample_size, radius, sample_shape, dimensions
                    )
                    boundary_correction = vol / score_vol
                    counts += (
                        len(tree.query_ball_point([x], radius)) - 1
                    ) / boundary_correction
                else:
                    counts += len(tree.query_ball_point([x], radius)) - 1
        elif dimensions == 2:
            score_vol = np.pi * radius**2
            if sample_shape == "circle":
                bound_size = np.pi * sample_size**2
            elif sample_shape == "rectangle":
                bound_size = sample_size[0] * sample_size[1]
            counts = 0
            for x, y in zip(d1, d2):
                if boundary_correct:
                    vol = calculate_overlap(
                        [x, y], sample_size, radius, sample_shape, dimensions
                    )
                    boundary_correction = vol / score_vol
                    counts += (
                        len(tree.query_ball_point([x, y], radius)) - 1
                    ) / boundary_correction
                else:
                    counts += len(tree.query_ball_point([x, y], radius)) - 1
        else:
            score_vol = (4 / 3) * np.pi * radius**3
            if sample_shape == "circle":
                bound_size = (4 / 3) * np.pi * sample_size**3
            elif sample_shape == "rectangle":
                bound_size = sample_size[0] * sample_size[1] * sample_size[2]
            counts = 0
            for x, y, z in zip(d1, d2, d3):
                if boundary_correct:
                    vol = calculate_overlap(
                        [x, y, z], sample_size, radius, sample_shape, dimensions
                    )
                    boundary_correction = vol / score_vol
                    counts += (
                        len(tree.query_ball_point([x, y, z], radius)) - 1
                    ) / boundary_correction
                else:
                    counts += len(tree.query_ball_point([x, y, z], radius)) - 1
        k_value = bound_size * counts / len(d1) ** 2
        if CSR_Normalise:
            k_value -= score_vol
        results.append(k_value)
    if len(results) == 1:
        return results[0]
    else:
        return results



import numpy as np
import pytest
from ripleyk import ripleyk
import random


@pytest.fixture(scope="module")
def random_points_3d():
    """
    Generates a set of random 3D points within a sphere of radius 1,
    based on the data generation code provided in the README.md.
    `random.seed(0)` is used for reproducibility.
    """
    xs = []
    ys = []
    zs = []
    radius = 1
    random.seed(0)
    # The loop runs 10,000 times, but the number of points generated
    # will be lower because some points fall outside the sphere.
    for _ in range(10000):
        positioned = False
        while not positioned:
            x = random.uniform(-radius, radius)
            y = random.uniform(-radius, radius)
            z = random.uniform(-radius, radius)
            if (x**2) + (y**2) + (z**2) < radius**2:
                xs.append(x)
                ys.append(y)
                zs.append(z)
                positioned = True
    return np.array(xs), np.array(ys), np.array(zs)


def test_calculate_ripley_2d_single_radius(random_points_3d):
    xs, ys, _ = random_points_3d
    radius = 0.5
    bounding_radius = 1
    k = ripleyk.calculate_ripley(radius, bounding_radius, d1=xs, d2=ys)
    assert np.isclose(k, 0.757289, atol=1e-6)


def test_calculate_ripley_2d_single_radius_boundary_correct(random_points_3d):
    xs, ys, _ = random_points_3d
    radius = 0.5
    bounding_radius = 1
    k = ripleyk.calculate_ripley(
        radius, bounding_radius, d1=xs, d2=ys, boundary_correct=True
    )
    assert np.isclose(k, 0.864648, atol=1e-6)


def test_calculate_ripley_2d_single_radius_boundary_correct_csr_normalise(
    random_points_3d,
):
    xs, ys, _ = random_points_3d
    radius = 0.5
    bounding_radius = 1
    k = ripleyk.calculate_ripley(
        radius,
        bounding_radius,
        d1=xs,
        d2=ys,
        boundary_correct=True,
        CSR_Normalise=True,
    )
    assert np.isclose(k, 0.079249, atol=1e-6)


def test_calculate_ripley_3d_multiple_radii(random_points_3d):
    xs, ys, zs = random_points_3d
    radii = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    expected_k = [
        -9.290137551123609e-06,
        -0.00010544767161566743,
        -0.00039142698870800463,
        -0.0007282757869681578,
        -0.0013863220751694216,
        -0.002301632731796177,
        -0.002895761209242842,
        -0.004294205083374969,
        -0.005929855937486295,
        -0.007915443959695345,
    ]
    k = ripleyk.calculate_ripley(
        radii,
        1,
        d1=xs,
        d2=ys,
        d3=zs,
        boundary_correct=True,
        CSR_Normalise=True,
    )
    assert np.allclose(k, expected_k, atol=1e-6)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ripleyk",
    version="0.9",
    author="Sam Ingram",
    author_email="sp_ingram12@yahoo.co.uk",
    description=("Calculation of the Ripley K (spatial statistics) value in python"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SamPIngram/RipleyK",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.19.2",
        "scipy>=1.5.2",
    ],
    python_requires=">=3.5",
)
